# encoding: utf-8
# frozen_string_literal: true
#
# PixCRUSH compression utility v1.0
# (C)2021 IRQsome Software
#

require 'chunky_png'

class ScronchPal
    
    attr_reader :i2c,:c2i

    def initialize(data)
        @i2c = []
        @c2i = {}
        data.each_line(chomp: true) do |line|
            if line =~ /\A(-?\d+) (-?\d+) (-?\d+)\Z/
                r = $1.to_i
                g = $2.to_i
                b = $3.to_i
                i = @i2c.size
                if (r>=0) && (g>=0) || (b>=0)
                    c = ChunkyPNG::Color.rgb(r,g,b)
                    @c2i[c] = i
                    @i2c << c
                elsif (r==-1) && (g==-1) && (b==-1)
                    @i2c << nil
                else 
                    raise
                end
            end
        end
    end
end


NTSC_PAL = ScronchPal.new(File.read("#{File.dirname(__FILE__)}/ntsc_scronch.pal"))
NTSC_LINES = 224
LINE_TERMINATOR = 0x07
RENDER_COGS = 3

def getpalidx(color)
    raise "wtf? #{color.inspect}" unless color.is_a? Integer
    NTSC_PAL.c2i[color] or raise "Color #{"%08X" % color} is not in palette"
end

def array2quad(a)
    #p "array2quad",a
    raise unless a.is_a? Array
    (
        (getpalidx(a[0])<<0) + 
        (getpalidx(a[1]||ChunkyPNG::Color::BLACK)<<8) + 
        (getpalidx(a[2]||ChunkyPNG::Color::BLACK)<<16) + 
        (getpalidx(a[3]||ChunkyPNG::Color::BLACK)<<24)
    )
end


def quad2array(q)
    [
        NTSC_PAL.i2c[(q>> 0)&255],
        NTSC_PAL.i2c[(q>> 8)&255],
        NTSC_PAL.i2c[(q>>16)&255],
        NTSC_PAL.i2c[(q>>24)&255],
    ]
end

def cmd_nil(length)
    raise unless length == 0
    return ''
end
def cmd_rle(length)
    raise if length > 64
    return (((length-1)<<2) + 0b10).chr
end
def cmd_raw8bpp(length)
    raise if length > 64
    return (((length-1)<<2) + 0b00).chr
end
def cmd_direct2bpp(length)
    raise if length > 64
    return (((length-1)<<2) + 0b01).chr
end
def cmd_new2bpp(length)
    return cmd_direct2bpp(length) if length == 1 # Avoid LINE_TERMINATOR
    raise if length > 32
    #puts "new2bpp #{length}"
    return (((length-1)<<3) + 0b111).chr
end
def cmd_current2bpp(length)
    raise if length > 32
    return (((length-1)<<3) + 0b011).chr
end

def find_extracolor(row,where,palette)
    # Not neccessarily the best way, but this is easy
    loop do
        return ChunkyPNG::Color::BLACK if where > row.length-16
        nsym = row[where,16]
        if nsym.uniq.count > 4 # Ignore areas that we will need to code as 8bpp, anyways
            where+=4
        else
            c = row[where,16].find{|c|!palette.include? c}
            return c if c
            where+=16
        end
    end
end

def next_encoding(row,where,palette,current_enc)
    remaining = row.length - where
    if remaining >= 4
        rlecol = row[where]
        rlelength = (row[where..].index{|c|c!=rlecol} || remaining) / 4 # rounding down...
        if rlelength >= 2 || (rlelength == 1 && remaining == 4) # Always do long RLE commands
            return :rle,palette,rlecol,rlelength # palette stays the same
        end
    end
    if remaining >= 16
=begin
        if remaining >= 32
            # Try 1bpp
            pal1 = row[where   ,16].uniq
            pal2 = row[where+16,16].uniq
            if pal1.count <= 2 && pal2.count <= 2
                # TODO 1bpp mode
            end
        end
=end
        # Try 2bpp with current palette
        if row[where,16].all?{|c|palette.include? c}
            return :current2bpp,palette # palette stays the same
        end
        if rlelength == 1 # Try short RLE
            return :rle,palette,rlecol,rlelength # palette stays the same
        end
        # Try 2bpp with new palette
        pal2bpp = row[where,16].uniq.sort
        if pal2bpp.count <= 4
            while pal2bpp.count < 4
                # Fill up palette.
                pal2bpp << find_extracolor(row,where,pal2bpp)
            end
            return :new2bpp,pal2bpp
        end
    end
    if remaining >= 4
        return :raw8bpp,palette
    else
        return :end,palette
    end
end

# Encode into 2bpp delta format
def encode2bpp(data,palette)
    tmp = 0
    pos = 0
    data.each do |c|
        idx = palette.index(c) or raise
        tmp = (tmp<<2)+((idx-pos)&3)
        pos = idx
    end
    return tmp
end

def encode_line(row)
    data,commands = "".b.dup,"".b.dup
    enccmd_histogram = {raw8bpp:0,new2bpp:0,current2bpp:0,direct2bpp:0,rle:0,nil:0}
    enclen_histogram = {raw8bpp:0,new2bpp:0,current2bpp:0,direct2bpp:0,rle:0,nil:0}
    raise "Width must be multiple of 16!" unless row.length % 16 == 0

    unless row.all?(ChunkyPNG::Color::BLACK)

        palette = quad2array(0x07_05_03_02)
        where = 0
        current_enc = :nil # either :raw8bpp, :new2bpp, :current2bpp or :direct2bpp
        enc_length = 0

        loop do
            # Figure out which encoding method we should apply
            #puts "@ #{where}"
            encsym,palette,rle_col,rle_length = next_encoding(row,where,palette,current_enc)
            #p encsym,palette,rle_col,rle_length
            case encsym
            when :rle
                commands << self.send(:"cmd_#{current_enc}",enc_length) # End current command
                enccmd_histogram[current_enc]+=1
                enclen_histogram[current_enc]+=enc_length
                # Put RLE command (RLE command is already as long as posible, so set current_enc to :nil)
                rletmp = rle_length
                while rletmp > 64
                    commands << cmd_rle(64)
                    commands << getpalidx(rle_col)
                    rletmp-=64
                end
                if rletmp > 0
                    commands << cmd_rle(rletmp)
                    commands << getpalidx(rle_col)
                end
                enccmd_histogram[:rle]+=1
                enclen_histogram[:rle]+=rle_length
                current_enc = :nil
                enc_length = 0
                where += rle_length*4
            when :new2bpp
                if current_enc == :direct2bpp || (current_enc == :new2bpp && enc_length == 1)
                    # Convert into :direct2bpp command
                    enc_length += 1
                    current_enc = :direct2bpp
                else
                    commands << self.send(:"cmd_#{current_enc}",enc_length) # End current command
                    enccmd_histogram[current_enc]+=1
                    enclen_histogram[current_enc]+=enc_length
                    # Start new 2bpp command
                    enc_length = 1
                    current_enc = :new2bpp
                end
                data << [array2quad(palette)].pack("V")
                data << [encode2bpp(row[where,16],palette)].pack("V")
                where += 16
            when :current2bpp
                if current_enc == :current2bpp || current_enc == :new2bpp
                    # append to current command
                    enc_length += 1
                else
                    commands << self.send(:"cmd_#{current_enc}",enc_length) # End current command
                    enccmd_histogram[current_enc]+=1
                    enclen_histogram[current_enc]+=enc_length
                    # Start new 2bpp command
                    enc_length = 1
                    current_enc = :current2bpp
                end
                data << [encode2bpp(row[where,16],palette)].pack("V")
                where += 16
            when :raw8bpp
                if current_enc == :raw8bpp
                    enc_length += 1
                else
                    commands << self.send(:"cmd_#{current_enc}",enc_length) # End current command
                    enccmd_histogram[current_enc]+=1
                    enclen_histogram[current_enc]+=enc_length
                    # Start new 8bpp command
                    enc_length = 1
                    current_enc = :raw8bpp
                end
                data << [array2quad(row[where,4])].pack("V")
                where += 4
            when :end
                break
            else raise end
        end
        commands << self.send(:"cmd_#{current_enc}",enc_length) # End current command
        enccmd_histogram[current_enc]+=1
        enclen_histogram[current_enc]+=enc_length
    end
    commands << LINE_TERMINATOR.chr
    enccmd_histogram.delete :nil
    enclen_histogram.delete :nil
    return data,commands,enccmd_histogram,enclen_histogram
end


if ARGV.size != 2 then puts <<~USAGE;exit; end
    PixCRUSH compression utility v1.0
    (C)2021 IRQsome Software
    Usage:
        ruby pixcrush.rb [input].png [output].dat
    
USAGE
incanv = ChunkyPNG::Image.from_file(ARGV[0])
databuffers = Array.new(RENDER_COGS){"".b.dup}
cmdbuffers = Array.new(RENDER_COGS){"".b.dup}


enccmd_histogram = {raw8bpp:0,new2bpp:0,current2bpp:0,direct2bpp:0,rle:0}
enclen_histogram = {raw8bpp:0,new2bpp:0,current2bpp:0,direct2bpp:0,rle:0}

incanv.height.times do |i|
    #puts "encoding line #{i}"
    linedata,linecmd,ch,lh = encode_line(incanv.row(i))
    databuffers[i%RENDER_COGS] << linedata
    cmdbuffers[i%RENDER_COGS] << linecmd
    enccmd_histogram.merge!(ch){|k,a,b|a+b}
    enclen_histogram.merge!(lh){|k,a,b|a+b}
end

# Pad with blank lines
blanklines = Rational(NTSC_LINES,RENDER_COGS).ceil
cmdbuffers.each{|b|b<<LINE_TERMINATOR.chr*blanklines}

dataoffsets = Array.new(RENDER_COGS){|i|databuffers[0...i].sum(&:length)+RENDER_COGS*4}
datatotal = databuffers.sum &:length
cmdoffsets = Array.new(RENDER_COGS){|i|cmdbuffers[0...i].sum(&:length)+datatotal+RENDER_COGS*4}
cmdtotal = cmdbuffers.sum &:length

ptrs = cmdoffsets.zip(dataoffsets).flatten.pack("v*")

File.open(ARGV[1],'wb') do |outfile|
    outfile.write(ptrs)
    databuffers.each{|b|outfile.write b}
    cmdbuffers.each{|b|outfile.write b}
end

puts "Done!"
puts "Generated #{cmdtotal} command bytes and #{datatotal} data bytes"
puts "Command histogram: #{enccmd_histogram.inspect}"
puts "Average symbols per command: #{enclen_histogram.merge(enccmd_histogram){|k,a,b|a.to_f / b}.inspect}"


